@@ -1,13 +1,13 @@
 #!/usr/bin/env python3
 #
-#
+# Stream-based realtime scientific data plotter.
 # Copyright (c) 2022, Hiroyuki Ohsaki.
 # All rights reserved.
 #
 
 import contextlib
-import time
 
+# Disable the banner when loading pygame modules.
 with contextlib.redirect_stdout(None):
     import pygame
     import pygame.gfxdraw
@@ -20,6 +20,7 @@ FONT_SIZE = 18
 
 BLACK = (0, 0, 0)
 
+# Internal color numbers.
 DARK_GRAY = 100
 GRAY = 101
 WHITE = 102
@@ -48,7 +49,13 @@ def hsv2rgb(h, s, v):
         return v, p, q
 
 class Series(list):
-    def __init__(self, id_=None, vmin=0, vmax=1, color=None, label=None):
+    def __init__(self,
+                 id_=None,
+                 vmin=0,
+                 vmax=1,
+                 color=None,
+                 label=None,
+                 hide=False):
         self.id_ = id_
         self.vmin = vmin
         self.vmax = vmax
@@ -58,12 +65,14 @@ class Series(list):
         if label is None:
             label = f'#{id_}'
         self.label = label
-        self.vsum = 0
+        self.hide = False
+        self.vsum = 0	# Used for obtaining the sample mean.
 
     def __repr__(self):
-        return f'Series(id_={self.id_}, vmin={self.vmin}, vmax={self.vmax}, color={self.color}, label={self.label}, len={len(self)}'
+        return f'Series(id_={self.id_}, vmin={self.vmin}, vmax={self.vmax}, color={self.color}, label={self.label}, hide={self.hide}, len={len(self)}'
 
     def append(self, v):
+        # Update the minimum and the maximum value at the time of data storage.
         if v < self.vmin:
             self.vmin = v
         if v > self.vmax:
@@ -72,12 +81,16 @@ class Series(list):
         super().append(v)
 
     def at(self, ratio, window=None):
+        """For a given window size WINDOW, pick a value located at the ratio
+        of RATIO in all samples withn the last WINDOW.  If window is not
+        specified, take a sample from all data."""
         if window is None:
             n = int(len(self) * ratio)
             return self[n]
+        # If the number of samples is not enough, pick from all available samples.
         if len(self) < window:
             return self.at(ratio, None)
-
+        # Otherwise, pick from the window of samples at the end.
         offset = len(self) - window
         n = int(offset + window * ratio)
         return self[n]
@@ -90,7 +103,7 @@ class Plot:
                  offset=None,
                  grid=None,
                  subgrid=None,
-                     start_color=0):
+                 start_color=0):
         self.screen = screen
         if screen:
             if width is None:
@@ -109,35 +122,44 @@ class Plot:
         self.window = None
         self.vmin = None
         self.vmax = None
-        self.excluded = []
 
     def __repr__(self):
         return f'Plot(grid={self.grid}, subgrid={self.subgrid}, _series={self._series})'
 
-    # series handling
+    # Series handling.
     def series(self, n):
+        """Return the object for N-th seriries.  The object is newly created
+        if it does not exist."""
         while len(self._series) - 1 < n:
             sr = Series(1 + n, color=self.start_color + n)
             self._series.append(sr)
         return self._series[n]
 
     def visible_series(self):
-        for sr in self._series:
-            if sr.id_ not in self.excluded:
-                yield sr
+        """Return all Seriries object, which are available as well as being
+        not hidden."""
+        return [sr for sr in self._series if len(sr) > 0 and not sr.hide]
 
-    def v2y(self, v):
+    def update_minmax(self, series):
+        """Find the maximum and the minimum values of all series."""
+        self.vmin = min([sr.vmin for sr in series])
+        self.vmax = max([sr.vmax for sr in series])
+
+    def to_y_coordinate(self, v):
+        """Assuming that the y-axis ranges from the minimum and the maximum
+        value, return the y-axsis of the value V.  For instance, if V is eqaul
+        to the minimum, this function returns the value of SELF.HEIGHT."""
         return int(self.height - self.height * (v - self.vmin) /
                    (self.vmax - self.vmin))
 
     def draw_single_series(self, sr):
-        if self.screen is None:
-            die('draw_single_series: no screen found')
+        """Draw a line for the seriries object SR as a concatenation of short
+        line segments."""
         x0, y0 = None, None
         for x in range(self.width):
             ratio = x / self.width
             v = sr.at(ratio, self.window)
-            y = self.v2y(v)
+            y = self.to_y_coordinate(v)
             if x > 0:
                 self.screen.draw_line(x0,
                                       y0,
@@ -147,24 +169,63 @@ class Plot:
                                       offset=self.offset)
             x0, y0 = x, y
 
-    def update_minmax(self, excluded=None):
-        self.vmin = min([sr.vmin for sr in self.visible_series()])
-        self.vmax = max([sr.vmax for sr in self.visible_series()])
+    def draw_grid(self, grid, color):
+        """Draw grids for y-axis.  The interval between grids is spcified by
+        GRID.  Its color is specified by COLOR."""
+        # How many grid lines should be drawn?
+        ngrids = (self.vmax - self.vmin) / grid
+        n = int(ngrids)
+        # Avoid too many lins in curses mode to avoid cluttering.
+        if self.screen.curses and n > 5:
+            return
+        frac = ngrids - n
+        v = self.vmin + grid * frac
+        for i in range(n):
+            y = self.to_y_coordinate(v)
+            # FIXME: It's better to control the brightntess with alpha channel.
+            self.screen.draw_line(0,
+                                  y,
+                                  self.width,
+                                  y,
+                                  color,
+                                  offset=self.offset)
+            v += grid
 
-    def draw_series(self, excluded=None):
-        self.update_minmax(excluded)
+    def draw_legends(self):
+        """Draw legends for all lines plotted."""
+        # NOTE: legends are shown also for hidden serries.
+        for n, sr in enumerate(self._series):
+            x = 1
+            y = 1 + n * FONT_SIZE
+            v = sr[-1]
+            # Display field label.
+            self.screen.draw_text(x, y, sr.label, sr.color, offset=self.offset)
+            # Display current, mean, and maximum values.
+            mean = sr.vsum / len(sr)
+            self.screen.draw_text(x + FONT_SIZE * 4,
+                                  y,
+                                  f'{v:8.2f}/{mean:8.2f}/{sr.vmax:8.2f}',
+                                  sr.color,
+                                  offset=self.offset)
+
+    def draw_series(self):
+        """Draw all series excluding hidden ones.  The range of the y-axis is
+        automatically determined."""
+        series = self.visible_series()
+        if not series:
+            return
+        self.update_minmax(series)
         # Draw grid line.
         if self.subgrid:
             self.draw_grid(self.subgrid, DARK_GRAY)
         if self.grid:
             self.draw_grid(self.grid, GRAY)
         # Draw all series.
-        for sr in self.visible_series():
+        for sr in series:
             self.draw_single_series(sr)
         # Draw legends.
-        # NOTE: legends are shown also for *excluded* serries
         self.draw_legends()
-        # Draw ticks.
+        # Display the maximum and the minimum values.
         self.screen.draw_text(self.width - FONT_SIZE * 6,
                               0,
                               f'{self.vmax:8.2f}',
@@ -176,39 +237,23 @@ class Plot:
                               WHITE,
                               offset=self.offset)
 
-    def draw_grid(self, grid, color):
-        ngrids = (self.vmax - self.vmin) / grid
-        n = int(ngrids)
-        if self.screen.curses and n > 5:
-            return
-        frac = ngrids - n
-        v = self.vmin + grid * frac
-        for i in range(n):
-            y = self.v2y(v)
+    def draw_background(self):
+        """Fill the background with gradient colors."""
+        # FIXME: Should not re-generate at every drawing.
+        for y in range(self.height):
+            alpha = 48 - int(48 * y / self.height)
             self.screen.draw_line(0,
                                   y,
                                   self.width,
                                   y,
-                                  color,
-                                  offset=self.offset)
-            v += grid
-
-    def draw_legends(self):
-        for n, sr in enumerate(self._series):
-            x = 1
-            y = 1 + n * FONT_SIZE
-            v = sr[-1]
-            self.screen.draw_text(x, y, sr.label, sr.color, offset=self.offset)
-            mean = sr.vsum / len(sr)
-            self.screen.draw_text(x + FONT_SIZE * 4,
-                                  y,
-                                  f'{v:8.2f}/{mean:8.2f}/{sr.vmax:8.2f}',
-                                  sr.color,
-                                  offset=self.offset)
+                                  self.start_color,
+                                  offset=self.offset,
+                                  alpha=alpha)
 
 class Screen:
     def __init__(self, curses=False, width=None, height=None):
         self.curses = curses
+        # Uses SVGA (800x600 pixels) window by default.
         if not curses and (width is None or height is None):
             width, height = 800, 600
         self.width = width
@@ -219,20 +264,24 @@ class Screen:
     def __repr__(self):
         return f'Screen(curses={self.curses}, width={self.width}, height={self.height})'
 
-    def color_rgb(self, n):
+    def color_rgba(self, n, alpha=255):
+        """Return 8-bit RGBA color for the color number N as a tuple.  The
+        alpha channel is given by ALPHA."""
         if n == DARK_GRAY:
             return 32, 32, 32
         if n == GRAY:
             return 64, 64, 64
         if n == WHITE:
             return 255, 255, 255
+        # Pick distinctive colors from HSV color space in this order.
         plist = [n / 10 for n in [6, 8, 1, 10, 0, 4, 9, 2, 5, 3, 7]]
         i = n % len(plist)
         r, g, b = hsv2rgb(255 * plist[i], .6, 1.)
-        return int(r * 255), int(g * 255), int(b * 255)
+        return int(r * 255), int(g * 255), int(b * 255), alpha
 
-    # draw primitives
+    # Draw primitives.
     def init_screen(self):
+        """Initialize the screen according to the output device."""
         if not self.curses:
             pygame.display.init()
             self.screen = pygame.display.set_mode((self.width, self.height))
@@ -241,7 +290,6 @@ class Screen:
             self.screen = curses.initscr()
             curses.start_color()
             self.height, self.width = self.screen.getmaxyx()
-            self.screen.erase()
             curses.init_pair(1, curses.COLOR_CYAN, curses.COLOR_BLACK)
             curses.init_pair(2, curses.COLOR_BLUE, curses.COLOR_BLACK)
             curses.init_pair(3, curses.COLOR_MAGENTA, curses.COLOR_BLACK)
@@ -249,22 +297,27 @@ class Screen:
             curses.init_pair(5, curses.COLOR_RED, curses.COLOR_BLACK)
             curses.init_pair(6, curses.COLOR_GREEN, curses.COLOR_BLACK)
             curses.init_pair(7, curses.COLOR_WHITE, curses.COLOR_BLACK)
+            self.screen.erase()
 
-    def draw_line(self, x1, y1, x2, y2, color=0, offset=None):
+    def draw_line(self, x1, y1, x2, y2, color=0, offset=None, alpha=255):
+        """Draw a straight line connecting two points: (X1, Y1) and (X2, Y2).
+        The line color can b especified by COLOR and ALPHA."""
         if offset is None:
             offset = 0, 0
         dx, dy = offset
         x1, y1 = x1 + dx, y1 + dy
         x2, y2 = x2 + dx, y2 + dy
         if not self.curses:
-            rgb = self.color_rgb(color)
-            pygame.gfxdraw.line(self.screen, x1, y1, x2, y2, rgb)
+            rgba = self.color_rgba(color, alpha)
+            pygame.gfxdraw.line(self.screen, x1, y1, x2, y2, rgba)
             return
 
         # Curses only.
+        # Change the point style for different line colors.
         points = ['●', '■', '◆', '▲', '▼', '◼', '○', '□', '◇', '△', '▽', '◻']
         point = points[color % len(points)]
 
+        # NOTE: These colors are only for drawing major and minor grids.
         if color == GRAY:
             point = '.'
             color = 1
@@ -272,6 +325,7 @@ class Screen:
             point = '-'
             color = 1
 
+        # A classical (and inefficient) line draing algorithm.
         if abs(x1 - x2) >= abs(y1 - y2):
             if x1 > x2:
                 x1, y1, x2, y2 = x2, y2, x1, y1
@@ -286,15 +340,18 @@ class Screen:
                 self.draw_text(x, y, point, color)
 
     def draw_text(self, x, y, text, color=0, size=FONT_SIZE, offset=None):
+        """Display a text TEXT at the location of (X, Y).  Font size and color
+        can be specified with COLOR and SIZE."""
         if offset is None:
             offset = 0, 0
         dx, dy = offset
         x, y = x + dx, y + dy
         if not self.curses:
-            font = pygame.font.SysFont('Courier', size)
-            rgb = self.color_rgb(color)
-            # Enable antialiasing.
-            text = font.render(text, True, rgb, BLACK)
+            # FIXME: Loaded font object should be re-utilized.
+            font = pygame.font.SysFont('Courier', size, bold=True)
+            rgba = self.color_rgba(color)
+            # The second parameter True means enabling antialiasing.
+            text = font.render(text, True, rgba, BLACK)
             self.screen.blit(text, (x, y))
         else:
             attr = curses.color_pair(1 + color % 7)
@@ -304,22 +361,26 @@ class Screen:
                 pass
 
     def clear(self):
+        """Clear the eintire screen."""
         if not self.curses:
             self.screen.fill(BLACK)
         else:
             self.screen.erase()
 
     def update(self):
+        """Update the screen to relfenct recent changes."""
         if not self.curses:
             pygame.display.update()
         else:
             self.screen.refresh()
 
     def wait(self):
+        """Wait until any key will be pressed."""
         if not self.curses:
             while True:
                 event = pygame.event.wait()
                 if event.type == pygame.KEYDOWN:
                     return
         else:
+            # Not necessary in curses mode.
             pass
